var preact = require('preact');

function createComponent(Store) {
    return /*@__PURE__*/(function (Component) {
        function Value(props, context) {
            Component.call(this, props);
            var value = props.value;
            var defaultValue = props.defaultValue;
            var store = props.store;
            var connected = !store;
            var controlled = false;
            if (!store) {
                if (value !== undefined) {
                    store = new Store(value, props);
                    controlled = true;
                } else {
                    store = new Store(defaultValue, props);
                }
            }
            this.state = {
                connected: connected,
                controlled: controlled,
                store: store,
                value: store.value
            };
        }

        if ( Component ) Value.__proto__ = Component;
        Value.prototype = Object.create( Component && Component.prototype );
        Value.prototype.constructor = Value;
        Value.prototype.componentDidMount = function componentDidMount () {
            var this$1 = this;

            var ref = this.state;
            var controlled = ref.controlled;
            var store = ref.store;
            this.mounted = true;
            this.unsubscribe = store.on(function (value) {
                if (!this$1.mounted) 
                    { return; }
                if (controlled) {
                    this$1.onChange(value);
                } else {
                    this$1.setState({
                        value: value
                    }, function () { return this$1.onChange(this$1.state.value); });
                }
            });
        };
        Value.prototype.componentWillUnmount = function componentWillUnmount () {
            this.mounted = false;
            this.unsubscribe();
        };
        Value.prototype.onChange = function onChange (value) {
            if (this.props.onChange) {
                this.props.onChange(value);
            }
        };
        Value.prototype.render = function render () {
            var ref = this.props;
            var children = ref.children;
            var render = ref.render;
            var disabled = ref.disabled; if ( disabled === void 0 ) disabled = false;
            var ref$1 = this.state;
            var controlled = ref$1.controlled;
            var connected = ref$1.connected;
            var store = ref$1.store;
            var value = controlled ? this.props.value : this.state.value;
            var transforms = disabled ? store.noops : store.transforms;
            var fn = children[0] || render;
            if (controlled) {
                store.value = value;
            }
            if (!connected) {
                store.props = this.props;
            }
            if (fn === null) {
                return null;
            }
            var renderProps = Object.assign({}, {value: value,
                disabled: disabled},
                transforms);
            for (var key in store.computeds) {
                renderProps[key] = store.computeds[key]();
            }
            var ret = typeof fn === 'function' ? fn(renderProps) : fn;
            return ret;
        };

        return Value;
    }(preact.Component));
}

function createFactory(Store, Value) {
    return function (initial, props) {
        var store = new Store(initial, props);
        function Connected(props) {
            return /*JSXElement?*/ preact.h( Value, Object.assign({}, props, { store: store }));
        }
        
        for (var k in store.transforms) {
            Connected[k] = store.transforms[k];
        }
        Connected.store = store;
        Connected.displayName = "Connected" + (Value.displayName);
        return Connected;
    };
}

var Store = function Store(value, props, empty) {
    var this$1 = this;
    if ( props === void 0 ) props = {};

    this.callbacks = [];
    this.value = value === undefined ? empty : value;
    this.transforms = {};
    this.noops = {};
    this.computeds = {};
    this.props = props || {};
    this.define('set', function (v, next) { return next; });
    this.define('reset', function () { return this$1.clone(value); });
    this.define('clear', function () { return this$1.clone(empty); });
    if (props.onChange) {
        this.on(props.onChange);
    }
};
Store.prototype.on = function on (callback) {
        var this$1 = this;

    var index = this.callbacks.push(callback) - 1;
    var off = function () { return this$1.callbacks[index] = null; };
    return off;
};
Store.prototype.clone = function clone (v) {
    return v;
};
Store.prototype.transform = function transform (fn, options) {
        if ( options === void 0 ) options = {};

    var ref = this;
        var value = ref.value;
        var callbacks = ref.callbacks;
    var mutates = options.mutates; if ( mutates === void 0 ) mutates = false;
    var current = mutates ? this.clone(value) : value;
    var next = typeof fn === 'function' ? fn(current) : fn;
    this.value = next;
    callbacks.forEach(function (cb) { return cb && cb(next); });
};
Store.prototype.define = function define (name, fn, options) {
        var this$1 = this;

    this.noops[name] = (function () {});
    this.transforms[name] = (function () {
            var args = [], len = arguments.length;
            while ( len-- ) args[ len ] = arguments[ len ];

        this$1.transform(function (v) { return fn.apply(void 0, [ v ].concat( args )); }, options);
    });
};
Store.prototype.compute = function compute (name, fn) {
        var this$1 = this;

    this.computeds[name] = (function () { return fn(this$1.value); });
};
Store.prototype.proxy = function proxy (method, options) {
        if ( options === void 0 ) options = {};

    var alias = options.alias; if ( alias === void 0 ) alias = method;
        var mutates = options.mutates; if ( mutates === void 0 ) mutates = false;
    this.define(alias, function (v) {
            var args = [], len = arguments.length - 1;
            while ( len-- > 0 ) args[ len ] = arguments[ len + 1 ];

        var ret = v[method].apply(v, args);
        return mutates ? v : ret;
    }, options);
};

var ArrayStore = /*@__PURE__*/(function (Store$$1) {
    function ArrayStore(value, props) {
        Store$$1.call(this, value, props, []);
        this.compute('first', function (v) { return v[0]; });
        this.compute('last', function (v) { return v[Math.max(0, v.length - 1)]; });
        this.proxy('concat');
        this.proxy('fill', {
            mutates: true
        });
        this.proxy('filter');
        this.proxy('flat');
        this.proxy('flatMap');
        this.proxy('map');
        this.proxy('pop', {
            mutates: true
        });
        this.proxy('push', {
            mutates: true
        });
        this.proxy('reverse', {
            mutates: true
        });
        this.proxy('shift', {
            mutates: true
        });
        this.proxy('slice');
        this.proxy('sort', {
            mutates: true
        });
        this.proxy('splice', {
            mutates: true
        });
        this.proxy('unshift', {
            mutates: true
        });
    }

    if ( Store$$1 ) ArrayStore.__proto__ = Store$$1;
    ArrayStore.prototype = Object.create( Store$$1 && Store$$1.prototype );
    ArrayStore.prototype.constructor = ArrayStore;
    ArrayStore.prototype.clone = function clone (value) {
        return value.slice();
    };

    return ArrayStore;
}(Store));

var BooleanStore = /*@__PURE__*/(function (Store$$1) {
    function BooleanStore(value, props) {
        Store$$1.call(this, value, props, false);
        this.define('toggle', function (v) { return !v; });
    }

    if ( Store$$1 ) BooleanStore.__proto__ = Store$$1;
    BooleanStore.prototype = Object.create( Store$$1 && Store$$1.prototype );
    BooleanStore.prototype.constructor = BooleanStore;

    return BooleanStore;
}(Store));

var SECONDS = 1000;
var MINUTES = 1000 * 60;
var HOURS = 1000 * 60 * 60;
var DateStore = /*@__PURE__*/(function (Store$$1) {
    function DateStore(value, props) {
        Store$$1.call(this, value, props, new Date());
        this.compute('date', function (v) { return v.getDate(); });
        this.compute('hours', function (v) { return v.getHours(); });
        this.compute('milliseconds', function (v) { return v.getMilliseconds(); });
        this.compute('minutes', function (v) { return v.getMinutes(); });
        this.compute('month', function (v) { return v.getMonth(); });
        this.compute('seconds', function (v) { return v.getSeconds(); });
        this.compute('year', function (v) { return v.getFullYear(); });
        this.define('setMonth', setMonth);
        this.proxy('setDate', {
            mutates: true
        });
        this.proxy('setFullYear', {
            alias: 'setYear',
            mutates: true
        });
        this.proxy('setFullYear', {
            mutates: true
        });
        this.proxy('setHours', {
            mutates: true
        });
        this.proxy('setMilliseconds', {
            mutates: true
        });
        this.proxy('setMinutes', {
            mutates: true
        });
        this.proxy('setSeconds', {
            mutates: true
        });
        this.define('incrementDate', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incDate(v, n);
        });
        this.define('incrementHours', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMil(v, n * HOURS);
        });
        this.define('incrementMilliseconds', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMil(v, n);
        });
        this.define('incrementMinutes', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMil(v, n * MINUTES);
        });
        this.define('incrementMonth', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMonth(v, n);
        });
        this.define('incrementSeconds', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMil(v, n * SECONDS);
        });
        this.define('incrementYear', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMonth(v, n * 12);
        });
        this.define('decrementDate', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incDate(v, 0 - n);
        });
        this.define('decrementHours', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMil(v, 0 - n * HOURS);
        });
        this.define('decrementMilliseconds', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMil(v, 0 - n);
        });
        this.define('decrementMinutes', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMil(v, 0 - n * MINUTES);
        });
        this.define('decrementMonth', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMonth(v, 0 - n);
        });
        this.define('decrementSeconds', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMil(v, 0 - n * SECONDS);
        });
        this.define('decrementYear', function (v, n) {
            if ( n === void 0 ) n = 1;

            return incMonth(v, 0 - n * 12);
        });
    }

    if ( Store$$1 ) DateStore.__proto__ = Store$$1;
    DateStore.prototype = Object.create( Store$$1 && Store$$1.prototype );
    DateStore.prototype.constructor = DateStore;
    DateStore.prototype.clone = function clone (value) {
        return new Date(value.getTime());
    };

    return DateStore;
}(Store));
function incMil(v, n) {
    return new Date(v.getTime() + n);
}

function incDate(v, n) {
    var d = new Date(v.getTime());
    d.setDate(d.getDate() + n);
    return d;
}

function incMonth(v, n) {
    var year = v.getFullYear();
    var desiredMonth = v.getMonth() + n;
    var desired = new Date(0);
    desired.setFullYear(year, desiredMonth, 1);
    desired.setHours(0, 0, 0, 0);
    var max = days(desired);
    v.setMonth(desiredMonth, Math.min(max, v.getDate()));
    return v;
}

function setMonth(v, m) {
    var d = new Date(v.getTime());
    var year = d.getFullYear();
    var day = d.getDate();
    var desired = new Date(0);
    desired.setFullYear(year, m, 15);
    desired.setHours(0, 0, 0, 0);
    var max = days(desired);
    d.setMonth(m, Math.min(day, max));
    return d;
}

function days(v) {
    var year = v.getFullYear();
    var monthIndex = v.getMonth();
    var lastDayOfMonth = new Date(0);
    lastDayOfMonth.setFullYear(year, monthIndex + 1, 0);
    lastDayOfMonth.setHours(0, 0, 0, 0);
    return lastDayOfMonth.getDate();
}

var MapStore = /*@__PURE__*/(function (Store$$1) {
    function MapStore(value, props) {
        Store$$1.call(this, value, props, new Map());
        this.define('set', function (v) {
            var a = [], len = arguments.length - 1;
            while ( len-- > 0 ) a[ len ] = arguments[ len + 1 ];

            var first = a[0];
            return a.length === 1 ? typeof first === 'function' ? first(v) : first : v.set.apply(v, a);
        });
        this.proxy('clear', {
            mutates: true
        });
        this.proxy('delete', {
            mutates: true
        });
        this.proxy('delete', {
            alias: 'unset',
            mutates: true
        });
    }

    if ( Store$$1 ) MapStore.__proto__ = Store$$1;
    MapStore.prototype = Object.create( Store$$1 && Store$$1.prototype );
    MapStore.prototype.constructor = MapStore;
    MapStore.prototype.clone = function clone (value) {
        return new Map(value);
    };

    return MapStore;
}(Store));

var NumberStore = /*@__PURE__*/(function (Store$$1) {
    function NumberStore(value, props) {
        var this$1 = this;

        Store$$1.call(this, value, props, 0);
        this.define('increment', function (v, n) {
            if ( n === void 0 ) n = 1;

            var num = typeof n !== 'number' ? 1 : n;
            return Math.min(v + num, this$1.props.max);
        });
        this.define('decrement', function (v, n) {
            if ( n === void 0 ) n = 1;

            var num = typeof n !== 'number' ? 1 : n;
            return Math.max(v - num, this$1.props.min);
        });
    }

    if ( Store$$1 ) NumberStore.__proto__ = Store$$1;
    NumberStore.prototype = Object.create( Store$$1 && Store$$1.prototype );
    NumberStore.prototype.constructor = NumberStore;

    return NumberStore;
}(Store));

var ObjectStore = /*@__PURE__*/(function (Store$$1) {
    function ObjectStore(value, props) {
        Store$$1.call(this, value, props, {});
        this.define('set', function (v) {
            var a = [], len = arguments.length - 1;
            while ( len-- > 0 ) a[ len ] = arguments[ len + 1 ];

            var first = a[0];
            if (a.length === 1) {
                return typeof first === 'function' ? first(v) : first;
            }
            var key = a[0];
            var val = a[1];
            var clone = Object.assign({}, v);
            clone[key] = val;
            return clone;
        });
        this.define('assign', function (v, val) {
            if ( val === void 0 ) val = {};

            return (Object.assign({}, v,
            val));
        });
        this.define('clear', function () { return ({}); });
        this.define('delete', unset);
        this.define('unset', unset);
    }

    if ( Store$$1 ) ObjectStore.__proto__ = Store$$1;
    ObjectStore.prototype = Object.create( Store$$1 && Store$$1.prototype );
    ObjectStore.prototype.constructor = ObjectStore;
    ObjectStore.prototype.clone = function clone (value) {
        return Object.assign({}, value);
    };

    return ObjectStore;
}(Store));
function unset(v, key) {
    var clone = Object.assign({}, v);
    delete clone[key];
    return clone;
}

var SetStore = /*@__PURE__*/(function (Store$$1) {
    function SetStore(value, props) {
        Store$$1.call(this, value, props, new Set());
        this.define('toggle', function (v, val, boolean) {
            if (boolean) {
                return v.add(val);
            } else {
                v.delete(val);
                return v;
            }
        }, {
            mutates: true
        });
        this.proxy('add');
        this.proxy('clear', {
            mutates: true
        });
        this.proxy('delete', {
            mutates: true
        });
        this.proxy('delete', {
            alias: 'remove',
            mutates: true
        });
    }

    if ( Store$$1 ) SetStore.__proto__ = Store$$1;
    SetStore.prototype = Object.create( Store$$1 && Store$$1.prototype );
    SetStore.prototype.constructor = SetStore;
    SetStore.prototype.clone = function clone (value) {
        return new Set(value);
    };

    return SetStore;
}(Store));

var StringStore = /*@__PURE__*/(function (Store$$1) {
    function StringStore(value, props) {
        Store$$1.call(this, value, props, '');
        this.proxy('concat');
        this.proxy('normalize');
        this.proxy('padEnd');
        this.proxy('padStart');
        this.proxy('repeat');
        this.proxy('replace');
        this.proxy('slice');
        this.proxy('substr');
        this.proxy('substring');
        this.proxy('toLowerCase');
        this.proxy('toUpperCase');
        this.proxy('trim');
        this.proxy('trimEnd');
        this.proxy('trimStart');
    }

    if ( Store$$1 ) StringStore.__proto__ = Store$$1;
    StringStore.prototype = Object.create( Store$$1 && Store$$1.prototype );
    StringStore.prototype.constructor = StringStore;

    return StringStore;
}(Store));

var Value = createComponent(Store);
var createValue = createFactory(Store, Value);
var ArrayValue = createComponent(ArrayStore);
var createArrayValue = createFactory(ArrayStore, ArrayValue);
var BooleanValue = createComponent(BooleanStore);
var createBooleanValue = createFactory(BooleanStore, BooleanValue);
var DateValue = createComponent(DateStore);
var createDateValue = createFactory(DateStore, DateValue);
var MapValue = createComponent(MapStore);
var createMapValue = createFactory(MapStore, MapValue);
var NumberValue = createComponent(NumberStore);
var createNumberValue = createFactory(NumberStore, NumberValue);
NumberValue.defaultProps = {
    max: Number.MAX_SAFE_INTEGER,
    min: Number.MIN_SAFE_INTEGER
};
var ObjectValue = createComponent(ObjectStore);
var createObjectValue = createFactory(ObjectStore, ObjectValue);
var SetValue = createComponent(SetStore);
var createSetValue = createFactory(SetStore, SetValue);
var StringValue = createComponent(StringStore);
var createStringValue = createFactory(StringStore, StringValue);

exports.Store = Store;
exports.Value = Value;
exports.createValue = createValue;
exports.ArrayStore = ArrayStore;
exports.ArrayValue = ArrayValue;
exports.createArrayValue = createArrayValue;
exports.BooleanStore = BooleanStore;
exports.BooleanValue = BooleanValue;
exports.createBooleanValue = createBooleanValue;
exports.DateStore = DateStore;
exports.DateValue = DateValue;
exports.createDateValue = createDateValue;
exports.MapStore = MapStore;
exports.MapValue = MapValue;
exports.createMapValue = createMapValue;
exports.NumberStore = NumberStore;
exports.NumberValue = NumberValue;
exports.createNumberValue = createNumberValue;
exports.ObjectStore = ObjectStore;
exports.ObjectValue = ObjectValue;
exports.createObjectValue = createObjectValue;
exports.SetStore = SetStore;
exports.SetValue = SetValue;
exports.createSetValue = createSetValue;
exports.StringStore = StringStore;
exports.StringValue = StringValue;
exports.createStringValue = createStringValue;
//# sourceMappingURL=preact.js.map
